/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.segmentation.activecontours.energies;

import java.util.Vector;

import de.unihalle.informatik.Alida.annotations.ALDClassParameter;
import de.unihalle.informatik.Alida.annotations.Parameter.ExpertMode;
import de.unihalle.informatik.Alida.exceptions.ALDException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.exceptions.MTBActiveContourException;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.exceptions.MTBActiveContourException.ExceptionType;
import de.unihalle.informatik.MiToBo.segmentation.basics.MTBSegmentationInterface;
import de.unihalle.informatik.MiToBo.segmentation.basics.CalcSegmentationStatistics;
import de.unihalle.informatik.MiToBo.segmentation.basics.CalcSegmentationStatistics.CalcTargets;

/**
 * Active contour energy implementing Chan-Vese region fitting.
 * <p>
 * This energy models the inner and outer regions of a single contour by a 
 * Gaussian model, i.e. each of the two regions is characterized by a mean 
 * intensity value, and deviations from this value are penalized. 
 * The energy is defined as follows:
 * {@latex.ilb %preamble{\\usepackage{amssymb}}
 *      \\begin{equation*}
 *      E(C) = \\lambda_{in} \\cdot 
 *       		\\int_{inside(C)} (I(x,y) - c_{in})^2 dx dy
 *       	+ \\lambda_{out} \\cdot  
 *       		\\int_{outside(C)} (I(x,y) - c_{out})^2 dx dy
 *      \\end{equation*}}
 * 
 * <p>
 * Note that class 0 is interpreted as background while class 1 is taken as 
 * foreground or contour interior. If there are additional labels present in 
 * the given segmentation they are ignored.
 * <p>
 * Paper: Chan and Vese, <i>Active Contours Without Edges</i>,
 * IEEE Transactions on Image Processing, vol. 10, no. 2, pp. 266-277, 2001.
 * 
 * @author moeller
 */
public class MTBActiveContourEnergy_CVRegionFit {

	/**
	 * Image to work on.
	 */
	@ALDClassParameter(label="Input Image", mode=ExpertMode.ADVANCED,
			dataIOOrder=-10)
	protected transient MTBImage inImg;
	
	/**
	 * Weighting factor for inner region fit.
	 */
	@ALDClassParameter(label="Interior Lambda",mode=ExpertMode.STANDARD,
			dataIOOrder=-5)
	protected double lambda_in = 0.5;

	/**
	 * Weighting factor for outer region fit.
	 */
	@ALDClassParameter(label="Exterior Lambda",mode=ExpertMode.STANDARD,
			dataIOOrder=-4)
	protected double lambda_out = 0.5;
	
	/**
	 * Operator to calculate energy parameters, i.e., some statistical 
	 * numbers on the regions of the segmentation.
	 */
	private transient CalcSegmentationStatistics imgStatsOp = null;
	
	/**
   * Region sizes.
   */
  protected transient int [] areas = null;

  /**
   * Average intensities.
   */
  protected transient double [] means = null;
  
	/**
   * Average intensities of non-regions.
   * <p>
   * A non-region is the complement of a region with regarding the segmentation,
   * i.e., subsumes all pixels belonging to any other region.
   */
  protected transient double [] means_compl = null;

  /**
   * Default constructor.
   */
  public MTBActiveContourEnergy_CVRegionFit() {
  	// nothing to do here
  }
  
  /**
   * Default constructor with arguments.
   * 
   * @param im	Image to work on.
   * @param l_in	Weighting factor for inner region fit.
   * @param l_out	Weighting factor for outer region fit.
   * @throws ALDOperatorException 
   */
  public MTBActiveContourEnergy_CVRegionFit(MTBImage im,
  			double lin, double lout) {
    this.inImg = im;
  	this.lambda_in= lin;
  	this.lambda_out= lout;
  }

  /**
   * Method to properly initialize instances of this class.
   * @throws ALDOperatorException
   */
  protected void init() throws MTBActiveContourException {
  	// do some savety checks
		if (this.inImg == null)
			throw new MTBActiveContourException(ExceptionType.INITIALIZATION_ERROR,
					"[MTBActiveContourEnergy_CVRegionFit] init() - no image given!");
		if (this.lambda_in < 0 || this.lambda_out < 0)
			throw new MTBActiveContourException(ExceptionType.INITIALIZATION_ERROR,
					"[MTBActiveContourEnergy_CVRegionFit] init() - " + 
							"at least one parameter is smaller than zero!");
  	// init the statistics operator
  	try {
	    this.imgStatsOp = new CalcSegmentationStatistics(this.inImg);
    } catch (ALDOperatorException e) {
			throw new MTBActiveContourException(ExceptionType.INITIALIZATION_ERROR,
					"[MTBActiveContourEnergy_CVRegionFit] init() - " + 
							"could not calculate segmentation statistics! " + 
									e.getCommentString());
    }
  	Vector<CalcTargets> toCalc= 
  		new Vector<CalcSegmentationStatistics.CalcTargets>();
  	toCalc.add(CalcTargets.classMean);
  	this.imgStatsOp.setTargets(toCalc);
  }
  
  /**
   * Updates parameters according to given membership.
   * @param mem		Current segmentation object.
   */
  protected void updateParameters(MTBSegmentationInterface mem) 
  	throws MTBActiveContourException {
  	
  	// update the energy parameter
  	try {
  		this.imgStatsOp.setSegmentation(mem);
	    this.imgStatsOp.runOp(true);
	    this.areas = this.imgStatsOp.getRegionSizes();
	  	this.means = this.imgStatsOp.getRegionMeans();
	  	this.means_compl = new double[this.means.length];
	  	
	  	double totalArea = 0;
	  	for (int i=0;i<this.areas.length;++i) 
	  		totalArea += this.areas[i];
	  	
	  	// update the complemental average values
	  	for (int i=0;i<this.means.length;++i) {
	  		double sum = 0;
	  		for (int j=0;j<this.means.length;++j) {
	  			if (j == i)
	  				continue;
	  			sum += this.means[j] * this.areas[j];
	  		}
	  		sum /= totalArea;
	  		this.means_compl[i] = sum;
	  	}
	  	
    } catch (ALDException e) {
    	throw new MTBActiveContourException(ExceptionType.UPDATE_ERROR, 
    		"[MTBActiveContourEnergy_CVRegionFit] " + 
    		"updateParameters() - update failed, keeping old parameters..." + 
    				e.getCommentString());
    }
  }

  /**
   * Returns the energy derivative value at position (x,y,z).
   */
  protected double getDerivative(
  	@SuppressWarnings("unused")MTBSegmentationInterface seg,int x,int y,int z){

  	// calculate statistical numbers of interior and exterior class
  	double mean_in = this.means[1];
  	double mean_out = this.means[0];
  	double pixInt = this.inImg.getValueDouble(x, y, z);
    return    this.lambda_in * (pixInt - mean_in) * (pixInt - mean_in)
    				- this.lambda_out * (pixInt - mean_out) * (pixInt - mean_out);
  }
  
  /**
   * Calculates the absolute energy value for the given segmentation.
   * @param seg	Segmentation to consider.
   * @return	Absolute value of energy.
   */
  protected double calcEnergy(MTBSegmentationInterface seg) {
  	// process the whole image
  	double [] classSums = new double[seg.getMaxLabel()+1];
    for (int z = 0; z < this.inImg.getSizeZ(); ++z) {
    	for (int y = 0; y < this.inImg.getSizeY(); ++y) {
    		for (int x = 0; x < this.inImg.getSizeX(); ++x) {

    			if (!seg.isVisible(x, y, z))
    				continue;

    			// check if image point lies inside or outside
    			int value = this.inImg.getValueInt(x, y, z);
    			int label = seg.getClass(x, y, z);

    			classSums[label] += 
    					(value - this.means[label]) * (value - this.means[label]); 
    		}
      }
    }
    double regionHomoTerm = 0.0;
  	regionHomoTerm += this.lambda_out * classSums[0];
  	regionHomoTerm += this.lambda_in * classSums[1];
    return regionHomoTerm;
  }
  
  /**
   * Calculates the interior part of energy value for the given segmentation.
   * @param seg	Segmentation to consider.
   * @return	Absolute value of interior energy.
   */
  protected double calcInteriorEnergy(MTBSegmentationInterface seg) {
  	// process the whole image
  	double [] classSums = new double[seg.getMaxLabel()+1];
    for (int z = 0; z < this.inImg.getSizeZ(); ++z) {
    	for (int y = 0; y < this.inImg.getSizeY(); ++y) {
    		for (int x = 0; x < this.inImg.getSizeX(); ++x) {

    			if (!seg.isVisible(x, y, z))
    				continue;

    			// check if image point lies inside or outside
    			int value = this.inImg.getValueInt(x, y, z);
    			int label = seg.getClass(x, y, z);

    			classSums[label] += 
    					(value - this.means[label]) * (value - this.means[label]); 
    		}
      }
    }
    return this.lambda_in * classSums[1];
  }

  /**
   * Returns the maximum possible derivative value this energy may yield. 
   * @return	Maximal derivative value.
   */
  protected double getEnergyDerivativeMaxVal() {
		return - Math.max(this.lambda_in, this.lambda_out);  	
  }
  
  /**
   * Returns the minimum possible derivative value this energy may yield. 
   * @return	Minimal derivative value.
   */
  protected double getEnergyDerivativeMinVal() {
		return - Math.min(this.lambda_in, this.lambda_out);
  }
  
  /**
   * Returns the average region intensities last calculated.
   * @return List of the regions' average intensity values.
   */
  protected double [] getMeans() {
  	return this.means.clone();
  }
  
	@Override
  public String toString() {
	  return "MTBActiveContourEnergy - CVRegionFit";
  }
}
